"""
Represent a player in the world and provide functions for monitoring their
state
"""
import math
import re
from time import sleep
from typing import List, Match, Pattern

from mcwb.types import Direction, Vec3
from mcwb.volume import Volume

from mciwb.threads import get_client

regex_coord = re.compile(r"\[(-?\d+.?\d*)d, *(-?\d+.?\d*)d, *(-?\d+.?\d*)d\]")
regex_angle = re.compile(r"-?\ *[0-9]+\.?[0-9]*(?:[Ee]\ *-?\ *[0-9]+)?")


class PlayerNotInWorld(Exception):
    pass


class Player:
    def __init__(self, name: str) -> None:
        self.name = name
        self._facing()

    def _get_entity_data(self, path: str, regex: Pattern[str]) -> Match[str]:
        """
        Get entity data with retries - the remote function sometimes fails to find
        an entity that does exist
        """
        client = get_client()
        for _ in range(5):
            data = client.data.get(entity=f"@e[name={self.name},limit=1]", path=path)
            match = regex.search(data)
            if match:
                return match
            sleep(0.1)

        raise PlayerNotInWorld(f"player {self.name} left")

    def _pos(self) -> Vec3:
        match = self._get_entity_data("Pos", regex_coord)
        pos = Vec3(
            float(match.group(1)), float(match.group(2)), float(match.group(3))
        ).with_ints()
        return pos

    @property
    def pos(self) -> Vec3:
        return self._pos()

    def _facing(self) -> Vec3:
        match = self._get_entity_data("Rotation", regex_angle)
        angle = float(match.group(0))

        index = int(((math.floor(angle) + 45) % 360) / 90)
        return Direction.cardinals[index]

    @property
    def facing(self) -> Vec3:
        return self._facing()

    def player_in(self, volume: Volume) -> bool:
        """
        Check if the player is in the volume
        """
        return volume.inside(self.pos)

    @classmethod
    def players_in(cls, volume: Volume) -> List["Player"]:
        """
        return a list of player names whose position is inside the volume
        """

        client = get_client()
        players = []

        names = [p.name for p in client.players.players]
        for name in names:
            try:
                pos = Player(name).pos
                if volume.inside(pos, 2):
                    players.append(name)
            except ValueError:
                pass  # players are sometimes missing temporarily
        return players



from demo.shapes import pyramid
from mciwb import Direction, Item, Monitor, Position, world


class Follower:
    def __init__(self, center=(0, 5, 0), size=11):
        self.center = center
        self.old_facing = None
        self.size = size
        self.pyramid_centre = Position(*center)
        half = int(size / 2)
        self.corner = (
            self.pyramid_centre + Direction.SOUTH * half + Direction.WEST * half
        )

    def follow(self):
        """
        Spooky Pyramid function: draws a pyramid centred at pyramid_centre, made
        of carved pumpkins that always turn to face the player
        """

        facing = Direction.facing(self.pyramid_centre, world.player.pos)
        if facing != self.old_facing:
            self.old_facing = facing
            pyramid(self.corner, self.size, Item.CARVED_PUMPKIN, facing=facing)


def follow_thread():
    # Use two threads for two pyramids
    p1 = Follower((0, 5, 0))
    p2 = Follower((20, 5, 0))
    p3 = Follower((0, 5, 20))
    p4 = Follower((20, 5, 20))
    Monitor([p1.follow, p2.follow, p3.follow, p4.follow])



"""
A mock client for testing functions that call the minecraft server.
This mocks a world of 100 blocks square with origin in the middle.
"""
from math import floor

import numpy as np
from mcipc.rcon.enumerations import CloneMode, Item, MaskMode
from mcwb.types import Items, Vec3
from mcwb.volume import Volume

from mciwb import Player


class MockClient:
    """
    A mock version of the RCON client for unit testing.
    """

    def __init__(self, host, port, passwd=None):
        # create an empty mock world full of air
        self.world = np.full((100, 100, 100), Item.AIR, dtype=Item)
        # note that numpy will wrap around indexes so this world has
        # ability to index from -50 to 50 in each dimension
        self.host = host
        self.port = port
        self.passwd = passwd

        self.connected = False

    ###########################################################################
    # the following are mock versions of the original Client Functions ########
    ###########################################################################

    def connect(self, retry=True):
        self.connected = True

    def setblock(self, pos: Vec3, block: Item):
        """set the block at position in the world"""
        self.world[pos.x, pos.y, pos.z] = block

    def fill(self, start: Vec3, end: Vec3, block: Item):
        block = Item(block)  # ensure enum
        self.world[
            floor(start.x) : floor(end.x),
            floor(start.y) : floor(end.y),
            floor(start.z) : floor(end.z),
        ] = block

    def clone(
        self,
        start_corner: Vec3,
        stop_corner: Vec3,
        dest: Vec3,
        mask_mode: MaskMode,
        clone_mode: CloneMode,
    ):
        """clone the contents of the world from a range to another position"""

        # use Volume to normalize the corners for use in slicing
        vol = Volume.from_corners(start_corner, stop_corner)

        v_start = vol.start
        v_stop = v_start + vol.size
        d_stop = dest + vol.size

        self.world[
            dest.x : d_stop.x, dest.y : d_stop.y, dest.z : d_stop.z
        ] = self.world[
            v_start.x : v_stop.x,
            v_start.y : v_stop.y,
            v_start.z : v_stop.z,
        ]
        return f"cloned {v_start} : {v_stop} to {dest}"

    @property
    def loot(self):
        """
        Support the 'grab' command which looks like this:
        == client.loot.spawn(dump).mine(vol.start + Vec3(*idx)) ==
        It is the convoluted way to find what block is at a coordinate.

        TODO: review this weird way to provide sub commands!
        """
        parent = self

        class spawn_cls:
            def spawn(self, dump):
                class mine_cls:
                    def mine(self, pos: Vec3):
                        return str(parent.getblock(pos))

                return mine_cls()

        return spawn_cls()

    def summon(self, entity_type, pos: Vec3, properties):
        """
        summon an entity at position with properties
        """
        # for simplicity, assume 'george' was summoned at 0,-16,-60, rotation=0
        # the data property below will deliver these values

    @property
    def players(self):
        class players_cls:
            @property
            def players(self):
                return [Player("george")]

        return players_cls()

    @property
    def data(self):
        class get_cls:
            def get(self, entity: str, path: str):
                if path == "Pos":
                    return "[0.0d, -60.0d, 0.0d]"
                elif path == "Rotation":
                    return "0"
                else:
                    return ""

        return get_cls()

    ###########################################################################
    # the following are additional functions for use in tests #################
    ###########################################################################

    def getblock(self, pos: Vec3) -> Item:
        """return the block at position in the world"""
        return self.world[int(pos.x), int(pos.y), int(pos.z)]

    def compare(self, pos: Vec3, cube: Items):
        """verify that the contents of the world at pos matches cube"""
        n_cube = np.array(cube)
        upper = pos + Vec3(*(n_cube.shape))
        world_slice = self.world[pos.x : upper.x, pos.y : upper.y, pos.z : upper.z]

        return np.array_equal(world_slice, n_cube)



import logging

from mciwb import Direction, Item, Position, world

logging.info("loading cool stuff")


def box(start: Position, size: int, item=Item.ACACIA_FENCE):
    """
    makes a cube of blocks with SW bottom corner at start

    Factored for easy understanding
    """

    # calculate the position of the NW bottom corner
    opposite = start + (Direction.EAST + Direction.NORTH) * (size - 1)

    # make rows of walls
    for up in range(size):
        # make a rows for each of the four walls
        for east in range(size):
            block_pos = start + Direction.EAST * east + Direction.UP * up
            world.set_block(block_pos, item)
            block_pos = opposite - Direction.EAST * east + Direction.UP * up
            world.set_block(block_pos, item)
        for north in range(size):
            block_pos = start + Direction.NORTH * north + Direction.UP * up
            world.set_block(block_pos, item)
            block_pos = opposite - Direction.NORTH * north + Direction.UP * up
            world.set_block(block_pos, item)

    for up in [0, size - 1]:
        for east in range(size):
            for north in range(size):
                block_pos = (
                    start
                    + Direction.NORTH * north
                    + Direction.EAST * east
                    + Direction.UP * up
                )
                world.set_block(block_pos, item)


def pyramid(start: Position, size: int, item=Item.COBBLESTONE, **kwargs):
    """
    Build a pyramid with South West Corner at 'start'
    """
    # make rows of walls
    row_size = size
    row_start = start
    for up in range(size):
        opposite = row_start + (Direction.EAST + Direction.NORTH) * (row_size - 1)
        # make a rows for each of the four walls
        for east in range(row_size):
            block_pos = row_start + Direction.EAST * east + Direction.UP * up
            world.set_block(block_pos, item, **kwargs)
            block_pos = opposite - Direction.EAST * east + Direction.UP * up
            world.set_block(block_pos, item, **kwargs)
        for north in range(row_size):
            block_pos = row_start + Direction.NORTH * north + Direction.UP * up
            world.set_block(block_pos, item, **kwargs)
            block_pos = opposite - Direction.NORTH * north + Direction.UP * up
            world.set_block(block_pos, item, **kwargs)

        row_start += Direction.EAST + Direction.NORTH
        row_size -= 2
        if row_size <= 0:
            break



"""
System tests for the Copy class
"""

import logging

from mcipc.rcon.item import Item
from mcwb.types import Anchor3, Vec3

from mciwb.copier import CopyPaste
from tests.cube import SampleCube

# TODO unit test version of this
# def test_world_reporting(mciwb_world: Iwb):
#     """
#     verify printing of the world object
#     """

#     assert f"player: {ENTITY_NAME}" in mciwb_world.__repr__()


def test_copy_anchors(mock_client):
    copy_anchors()


def copy_anchors():
    """
    Test copy and paste of the same cube using all possible opposite corners
    for a total of 8 pairs of opposite corner pairs (with ordering significant)

    Because the start corner is the anchor for the paste, every paste will
    place the cube offset in a different direction
    """

    copier = CopyPaste()

    t = SampleCube()
    source = Vec3(2, 5, 2)
    t.create(source)

    dest = Vec3(10, 5, 2)

    # try anchoring to each of the 8 corners of the cube
    corner_pairs = (
        ((2, 5, 2), (4, 7, 4), Anchor3.BOTTOM_NW),
        ((4, 5, 2), (2, 7, 4), Anchor3.BOTTOM_NE),
        ((2, 7, 2), (4, 5, 4), Anchor3.TOP_NW),
        ((2, 5, 4), (4, 7, 2), Anchor3.BOTTOM_SW),
        ((4, 7, 2), (2, 5, 4), Anchor3.TOP_NE),
        ((2, 7, 4), (4, 5, 2), Anchor3.TOP_SW),
        ((4, 5, 4), (2, 7, 2), Anchor3.BOTTOM_SE),
        ((4, 7, 4), (2, 5, 2), Anchor3.TOP_SE),
    )
    try:

        for start, stop, anchor in corner_pairs:
            logging.info("copy anchors test: %s, %s, %s", start, stop, anchor)
            copier.select(Vec3(*stop))
            copier.select(Vec3(*start))

            copier.paste_safe(Vec3(*dest))

            try:
                t.test(dest, anchor)
            finally:
                t.clear(dest, anchor)
    finally:
        t.clear(source)


def test_expand(mock_client):
    copier = CopyPaste()

    start = Vec3(0, 0, 0)
    stop = Vec3(5, 5, 5)
    copier.select(stop)
    copier.select(start)

    copier.expand(x=1, z=2)
    assert copier.start_pos == Vec3(0, 0, 0)
    assert copier.stop_pos == Vec3(6, 5, 7)

    copier.expand(x=-1, y=-2)
    assert copier.start_pos == Vec3(-1, -2, 0)
    assert copier.stop_pos == Vec3(6, 5, 7)

    exp = Vec3(0, 20, 20)
    copier.expand_to(exp)
    assert copier.start_pos == Vec3(-1, -2, 0)
    assert copier.stop_pos == Vec3(6, 20, 20)

    exp = Vec3(-20, -20, -20)
    copier.expand_to(exp)
    assert copier.start_pos == Vec3(-20, -20, -20)
    assert copier.stop_pos == Vec3(6, 20, 20)


def test_clear(mock_client):
    copier = CopyPaste()

    mock_client.setblock(Vec3(0, 0, 0), Item.STONE)
    copier.select(Vec3(1, 1, 1))
    copier.select(Vec3(0, 0, 0))

    copier.clear()

    assert mock_client.getblock(Vec3(0, 0, 0)) == Item.AIR



import logging
from time import sleep

import pytest
from docker.models.containers import Container
from mcipc.rcon.enumerations import Item
from mcipc.rcon.je import Client

from mciwb.iwb import Iwb
from mciwb.player import Player
from mciwb.threads import get_client, set_client
from tests.mockclient import MockClient
from tests.server import (
    ENTITY_NAME,
    ENTITY_POS,
    HOST,
    RCON_P,
    RCON_PORT,
    MinecraftServer,
)

logging.basicConfig(
    level=logging.DEBUG,
    format="%(levelname)s: %(funcName)s - %(message)s (%(pathname)s:%(lineno)d)",
)


@pytest.fixture(scope="session")
def minecraft_container(request: pytest.FixtureRequest):
    """
    Spin up a test minecraft server and return a container object

    This fixture is run once per session, meaning that it does not need
    to be executed twice for two runs of the tests. This requires
    caution as the world must be reset to a known state.
    """
    mc = MinecraftServer(name="mciwb_server", rcon=RCON_PORT)
    mc.minecraft_create()

    request.addfinalizer(mc.minecraft_remove)
    return mc


def client_connect():
    """
    Make an rcon connection, waiting for the server to come online
    """
    client = Client(HOST, RCON_PORT, passwd=RCON_P)

    for _ in range(10):
        try:
            client.connect(True)
        except ConnectionRefusedError:
            sleep(1)
        else:
            break
    else:
        raise TimeoutError("cannot connect to Minecraft Server")

    return client


@pytest.fixture(scope="session")
def minecraft_client(minecraft_container: Container):
    """
    return an rcon connection to the test server
    """
    client = client_connect()

    # attach the new client to the current thread
    set_client(client)

    # ensure that the chunks near 0,0,0 are loaded
    client.setworldspawn(ENTITY_POS)
    # don't announce every rcon command
    client.gamerule("sendCommandFeedback", False)

    # make sure the local chunk is loaded even if we don't summon a player
    client.forceload.add((0, 0), (0, 0))
    # make sure that the grab function entities that are created as a side
    # effect will drop into the void
    # TODO in 1.19 the bottom of the world is at -64 so need a fix for that!
    client.setblock((0, 0, 0), Item.AIR.value)

    return client


@pytest.fixture()
def minecraft_player():
    client = get_client()

    # summon a fixed position, named entity as a substitute for a player
    res = client.summon("armor_stand", ENTITY_POS, {"CustomName": f'"{ENTITY_NAME}"'})
    logging.info(res)

    for _ in range(10):
        try:
            player = Player(ENTITY_NAME)
        except ValueError:
            sleep(1)
        else:
            break
    else:
        raise ValueError("dummy player creation failed")

    return player


@pytest.fixture()
def mciwb_world(
    request, minecraft_container, minecraft_client: Client, minecraft_player: Player
):
    def stop_thread():
        world.stop()

    request.addfinalizer(stop_thread)

    world = Iwb(HOST, RCON_PORT, passwd=RCON_P)
    world.add_player(ENTITY_NAME)

    return world


@pytest.fixture()
def mock_client():
    client = MockClient("localhost", 20400, "pass")
    set_client(client)  # type: ignore
    return client



"""
System tests for the Copy class
"""


from time import sleep

from mcipc.rcon.item import Item

from mciwb import Direction, Iwb
from mciwb.player import Player
from mciwb.threads import get_client
from tests.conftest import ENTITY_NAME


def test_session_fixtures(mciwb_world: Iwb, minecraft_player: Player):
    """
    verify that the global session fixtures that create a minecraft server,
    RCON client and mciwb.Player are working
    """
    assert mciwb_world.player.pos == minecraft_player.pos


def test_world_reporting(mciwb_world: Iwb):
    """
    verify printing of the world object
    """
    assert f"player: {ENTITY_NAME}" in mciwb_world.__repr__()


def test_world_player_signs(mciwb_world: Iwb, minecraft_player: Player):
    """
    verify printing of the world object
    """
    sp = 'minecraft:oak_sign{{Text1:\'{{"text":"' "{0}" """"}}'}}"""

    client = get_client()
    # place some action signs in the world and verify they are actioned
    select = sp.format("select")
    extend = sp.format("extend")
    paste = sp.format("paste")

    p = minecraft_player.pos
    client.setblock(p + Direction.SOUTH * 3, select)
    sleep(0.5)
    client.setblock(p + Direction.SOUTH * 3, select)
    sleep(0.5)
    client.setblock(p + Direction.SOUTH * 2 + Direction.UP, extend)
    sleep(0.5)
    client.setblock(p + Direction.SOUTH * 2 + Direction.UP, paste)
    sleep(0.5)

    block = mciwb_world.get_block(p + Direction.SOUTH * 2)

    assert block == Item.GRASS_BLOCK

    mciwb_world.set_block(p + Direction.SOUTH * 2, Item.AIR)



import logging
import os
import shutil
from datetime import datetime
from pathlib import Path
from typing import Optional
from zipfile import ZIP_DEFLATED, ZipFile

from mciwb.threads import get_client


class Backup:
    def __init__(
        self,
        world_name: str,
        world_folder: str,
        backup_folder: str,
    ):
        self.name = world_name
        self.world_folder = Path(world_folder)
        self.backup_folder = Path(backup_folder)

        if not self.backup_folder.exists():
            raise ValueError("backup folder must exist")
        if not (self.world_folder / "level.dat"):
            raise ValueError(f"{world_folder} does not look like a minecraft world")

    def backup(self, running=True):
        client = get_client()

        if self.name == "" or client is None:
            logging.error("No backup details available.")
            return

        fname = datetime.strftime(datetime.now(), f"%y-%m-%d.%H.%M.%S-{self.name}.zip")

        client.say(f"Preparing to backup to {fname}")
        result = client.save_off()
        logging.debug("save_off: " + result)

        result = client.save_all()
        logging.debug("save_all: " + result)
        client.say("Backing up ...")

        file = self.backup_folder / fname
        world_files = self.world_folder.glob("**/*")
        world_files = [f for f in world_files if not f.suffix == ".lock"]
        with ZipFile(file, "w", compression=ZIP_DEFLATED) as zip_file:
            for wf in world_files:

                zip_file.write(wf, arcname=wf.relative_to(self.world_folder))
        logging.debug("ZipFile complete")

        result = client.save_on()
        logging.debug("save_on: " + result)
        client.say("Backup Complete.")

    def get_latest_zip(self) -> Path:
        backups = self.backup_folder.glob("*.zip")
        ordered = sorted(backups, key=os.path.getctime, reverse=True)
        return ordered[0]

    def restore(self, fname: Optional[Path] = None, backup=False):

        """
        restore world from backup. This must be called with the server stopped.
        """

        if not fname:
            fname = self.get_latest_zip()
        if not fname.exists():
            raise ValueError("{file} not found")

        # backup for recovery from accidental recovery ! Note that on some filesystems
        # it is OK to do this while the server is running. On other filesystems the
        # server will need to be stopped before calling this function
        if backup:
            old_world = Path(str(self.world_folder) + "-old")
            if old_world.exists():
                shutil.rmtree(old_world)
            shutil.move(str(self.world_folder), str(old_world))

        # restore zipped up backup to world folder
        with ZipFile(fname, "r") as zip_file:
            zip_file.extractall(path=self.world_folder)

        # remove lockfile if it exists
        for file in self.world_folder.glob("*.lock"):
            logging.debug(f"removing {file}")
            file.unlink()

        logging.info(f"Restored {self.world_folder} from {fname}")



"""
Functions for launching and controlling a Minecraft server in a Docker container.
"""
import logging
import os
import shutil
from datetime import datetime
from pathlib import Path
from tempfile import gettempdir
from time import sleep

from docker import from_env
from docker.models.containers import Container
from mcwb.types import Vec3

from mciwb import Client

KEEP_SERVER = "MCIWB_KEEP_SERVER" in os.environ

HOST = "localhost"
RCON_PORT = 20401
RCON_P = "pass"
ENTITY_NAME = "george"
ENTITY_POS = Vec3(0, -60, 0)

# the locally mapped temporary folder for minecraft data

data_folder = Path(gettempdir()) / "test_mc_servers"


class MinecraftServer:
    def __init__(self, name: str, rcon: int) -> None:
        self.rcon = rcon
        self.name = name
        self.folder = data_folder / name
        self.world = self.folder / "world"
        self.cont = None

    def wait_server(self):
        """
        Wait until the server is ready to accept rcon connections

        Multiple calls to this with container restarts require passing a count.
        This is how many times the wait code looks for the server to come online
        in the logs.

        This is required because the --since argument to the docker logs command
        fails to return any logs at all.
        """

        start_time: datetime = datetime.now()
        timeout = 100

        assert isinstance(self.cont, Container)

        self.cont.reload()
        if self.cont.status != "running":
            logs = "\n".join(str(self.cont.logs()).split(r"\n"))
            raise RuntimeError(f"minecraft server failed to start\n\n{logs}")

        logging.info("waiting for mc server to come online")
        for block in self.cont.logs(stream=True):
            logging.debug(block.decode("utf-8").strip())
            if b"RCON running" in block:
                break
            elapsed = datetime.now() - start_time
            if elapsed.total_seconds() > timeout:
                raise RuntimeError("Timeout Starting minecraft")

        # wait until a connection is available
        for _ in range(10):
            try:
                with Client(HOST, self.rcon, passwd=RCON_P):
                    pass
                break
            except ConnectionRefusedError:
                sleep(2)
        else:
            raise RuntimeError("Timeout Starting minecraft")
        logging.info("mc server is online")

    def stop(self):
        """
        Stop the minecraft server
        """
        assert isinstance(self.cont, Container)
        logging.info("Stopping Minecraft Server {self.name} ...")
        self.cont.stop()
        self.cont.wait()

        logging.info("Stopped Minecraft Server {self.name} ...")

    def start(self):
        """
        Start the minecraft server
        """
        assert isinstance(self.cont, Container)
        logging.info("Starting Minecraft Server {self.name} ...")
        self.cont.start()
        self.wait_server()
        logging.info("Started Minecraft Server {self.name} ...")

    def minecraft_remove(self):
        """
        Remove a minecraft server container
        """
        # set env var MCIWB_KEEP_SERVER to keep server alive for faster
        # repeated tests and viewing the world with a minecraft client
        if self.cont and not KEEP_SERVER:
            logging.info("Removing Minecraft Server {self.name} ...")
            self.stop()
            self.cont.remove()

    def minecraft_create(self, world=None) -> None:
        """
        Spin up a test minecraft server and return a container object

        This fixture is run once per session, meaning that it does not need
        to be executed twice for two runs of the tests. This requires
        caution as the world must be reset to a known state.
        """

        # create and launch minecraft container once per session
        docker_client = from_env()

        for container in docker_client.containers.list(all=True):
            assert isinstance(container, Container)
            if container.name == self.name:
                self.cont = container
                if container.status == "running":
                    logging.info(f"test minecraft server '{self.name}' already running")
                    return
                else:
                    logging.info(
                        f"test minecraft server '{self.name}' exists. restarting"
                    )
                    container.start()
                    self.wait_server()
                    return

        env = {
            "EULA": "TRUE",
            # "VERSION": "1.17.1",
            "SERVER_PORT": self.rcon + 1,
            "RCON_PORT": self.rcon,
            "ENABLE_RCON": "true",
            "RCON_PASSWORD": RCON_P,
            "GENERATE_STRUCTURES": "false",
            "SPAWN_ANIMALS": "false",
            "SPAWN_MONSTERS": "false",
            "SPAWN_NPCS": "false",
            "VIEW_DISTANCE": " 5",
            "SEED": 0,
            "LEVEL_TYPE": "FLAT",
            "OPS": "TransformerScorn",
            "MODE": "creative",
            "SPAWN_PROTECTION": "FALSE",
        }
        # offline mode disables OPS so dont use it if we are keeping the server
        # for local testing. But normally for running CI we want this option.
        if not KEEP_SERVER:
            env["ONLINE_MODE"] = "FALSE"
        if world:
            env["WORLD"] = str(world)

        if not self.folder.exists():
            self.folder.mkdir(parents=True)
        else:
            shutil.rmtree(self.folder)
            self.folder.mkdir(parents=True)

        container = docker_client.containers.run(
            "itzg/minecraft-server",
            detach=True,
            environment=env,
            network_mode="host",
            restart_policy={"Name": "no"},
            volumes={
                str(self.folder): {"bind": "/data", "mode": "rw"},
                str(data_folder): {"bind": str(data_folder), "mode": "rw"},
            },
            name=self.name,
        )

        self.cont = container

        self.wait_server()



import logging
import os
import shutil
from pathlib import Path

import pytest

# import pytest
from mcipc.rcon.item import Item
from mcwb.itemlists import grab
from mcwb.types import Vec3
from mcwb.volume import Volume

from mciwb import Client
from mciwb.backup import Backup
from mciwb.threads import set_client
from tests.conftest import HOST, RCON_P
from tests.server import KEEP_SERVER, MinecraftServer, data_folder

GITHUB_ACTIONS = "GITHUB_ACTIONS" in os.environ


@pytest.mark.skipif(KEEP_SERVER, reason="incompatible with keep server")
def test_backup_restore(tmp_path: Path):
    """
    Test the backup and restore functionality.

    Backup a server, make a change to the world, restore the backup, and
    verify that the world has been restored.

    Requires use of client contexts so that we can restart clients.

    (TODO: need to look at why client.close does not work, as contexts were
    the only way to get this working)
    """

    BACKUP_RCON = 20600
    backup_folder = data_folder / "backup"
    if backup_folder.exists():
        shutil.rmtree(backup_folder)
    backup_folder.mkdir(parents=True)

    test_block = Vec3(100, 100, 100)

    # create a new world to backup from
    mc_backup = MinecraftServer(name="mciwb_backup", rcon=BACKUP_RCON)
    backup = Backup("backup_world", str(mc_backup.world), str(backup_folder))
    mc_backup.minecraft_create()

    # make a change to the world which is to be backed up
    with Client(HOST, BACKUP_RCON, passwd=RCON_P) as client:
        set_client(client)
        result = client.setblock(test_block, Item.RED_CONCRETE.value)
        logging.info("setblock %s", result)
        # backup the world
        backup.backup()

    RESTORE_RCON = 20500
    # create a new world to restore into
    mc_restore = MinecraftServer(name="mciwb_restore", rcon=RESTORE_RCON)
    if GITHUB_ACTIONS:
        # GHA has issues with file locks on the stopped container
        # use mc docker's feature of creating from zip instead
        mc_restore.minecraft_create(world=backup.get_latest_zip())
    else:
        mc_restore.minecraft_create()
        mc_restore.stop()
        restore = Backup("restore_world", str(mc_restore.world), str(backup_folder))
        # restore the world and the start the server
        restore.restore()
        mc_restore.start()

    with Client(HOST, RESTORE_RCON, passwd=RCON_P) as client:
        set_client(client)
        # TODO mcwb should break out a getblock function from grab
        grab_volume = Volume.from_corners(test_block, test_block)
        restored_blocks = grab(client, grab_volume)

    # shut down the servers
    mc_restore.minecraft_remove()
    mc_backup.minecraft_remove()

    assert restored_blocks[0][0][0] == Item.RED_CONCRETE



from mcipc.rcon.item import Item
from mcipc.rcon.je import Client
from mcwb import Anchor3 as Corner
from mcwb import Blocks, Cuboid
from mcwb import Vec3 as Position
from mcwb.types import Direction, Planes3d

from mciwb.copier import CopyPaste
from mciwb.iwb import Iwb, world
from mciwb.monitor import Monitor
from mciwb.player import Player

try:
    # Use live version from git
    from setuptools_scm import get_version

    # Warning: If the install is nested to the same depth, this will always succeed
    tmp_version = get_version(root="../../", relative_to=__file__)
    del get_version
except (ImportError, LookupError):
    # Use installed version
    from ._version import version as __version__  # type: ignore
else:
    __version__ = tmp_version

# imports here are used to bring commonly used classes into a single namespace
# for ease of use for novice programmers
__all__ = [
    "__version__",
    "Item",
    "Iwb",
    "Client",
    "Corner",
    "Blocks",
    "Cuboid",
    "Position",
    "Direction",
    "Planes3d",
    "CopyPaste",
    "Monitor",
    "Player",
    "world",
]



"""
Add an interactive capability through the placing of signs in the world
"""
import logging
import re
from time import sleep
from typing import Dict

from mcwb.types import Item, Vec3

from mciwb.copier import CopyPaste
from mciwb.monitor import CallbackPosFunction
from mciwb.player import Player
from mciwb.threads import get_client


class Signs:
    """
    Each sign object can be used to monitor the placing of signs by a single
    player. The object can be hooked to functions that are to be called when
    a sign is placed by the player.

    By default each Sign object is initialized with select/copy/paste
    functions
    """

    re_sign_text = re.compile(r"""Text1: '{"text":"([^"]*)"}'""")
    re_sign_entity = (
        """minecraft:oak_sign{{BlockEntityTag:{{Text1:'{{"text":"{0}"}}'}},"""
        """display:{{Name:'{{"text":"{0}"}}'}}}}"""
    )

    wall_sign = "minecraft:oak_wall_sign"

    def __init__(self, player: Player):
        self.player = player
        self.copy = CopyPaste()
        self.signs: Dict[str, CallbackPosFunction] = self.copy.get_commands()

    def get_target_block(self, pos: Vec3, facing: Vec3) -> Vec3:
        """
        determine the target block that the sign at pos indicates
        """
        # use 'execute if' with a benign command like seed
        result = get_client().execute.if_.block(pos, self.wall_sign).run("seed")

        if "Seed" in result:
            # wall signs target the block behind them
            pos += facing
        else:
            # standing signs target the block below them
            pos += Vec3(0, -1, 0)

        return pos

    def poll(self):
        """
        check if a sign has been placed in front of the player
        1 to 4 blocks away and take action based on sign text
        """
        client = get_client()

        facing = self.player.facing
        player_pos = self.player.pos
        for height in range(-1, 3):
            for distance in range(1, 4):
                sleep(0)  # don't hog the connection
                pos = player_pos + facing * distance
                block_pos = pos.with_ints() + Vec3(0, height, 0)
                data = client.data.get(block=block_pos)
                match = self.re_sign_text.search(data)
                if match:
                    text = match.group(1)
                    logging.debug(f"Sign at {pos} has text {text}")
                    target = self.get_target_block(block_pos, facing)
                    client.setblock(
                        block_pos,
                        Item.AIR,  # type: ignore
                    )
                    self.do_action(text, target)

    def do_action(self, command: str, target: Vec3):
        # if the command is not found then this is just an ordinary sign (I assume!)
        if command in self.signs:
            logging.debug(f"Calling action for {command}")
            self.signs[command](target)

    def add_sign(self, name: str, function: CallbackPosFunction):
        self.signs[name] = function

    def remove_sign(self, name: str, function: CallbackPosFunction):
        del self.signs[name]

    def give_signs(self):
        """
        Give player one of each command sign in our commands list
        """
        client = get_client()
        for command in self.signs:
            client.give(self.player.name, self.re_sign_entity.format(command))



"""
Thread functions for monitoring state of the world
"""

import logging
from time import sleep
from typing import Callable, List, Union

from mcwb import Vec3
from rcon.exceptions import SessionTimeout

from mciwb.player import PlayerNotInWorld
from mciwb.threads import get_client, get_thread_name, new_thread

CallbackFunction = Callable[[], None]
CallbackPosFunction = Callable[[Vec3], None]


class Monitor:
    """ "
    A class to provide threads for monitoring. Each thread maintains a
    list of functions to call and its own Client object for parallel
    execution of Minecraft server functions.
    """

    monitor_num = 0
    monitors: List["Monitor"] = []

    def __init__(
        self,
        func: Union[None, CallbackFunction, List[CallbackFunction]] = None,
        once=False,
        name=None,
        poll_rate=0.2,
    ) -> None:

        if name is None:
            name = f"Monitor{Monitor.monitor_num}"
            Monitor.monitor_num += 1

        self.pollers: List[CallbackFunction] = (
            [] if func is None else func if isinstance(func, list) else [func]
        )

        self.name = name
        self.once = once
        self.poll_rate = poll_rate
        self.poll_thread = None

        self._start_poller()

    def _start_poller(self):
        if self.poll_thread is None:
            logging.debug(f"starting polling thread {self.name}")
            self.poll_thread = new_thread(get_client(), self._poller, self.name)
            self.monitors.append(self)
            self._polling = True

    def _poller(self):
        """
        the polling function will run until the monitor is stopped
        """

        while self._polling:
            try:
                for func in self.pollers:
                    func()
                sleep(self.poll_rate)
            except BrokenPipeError:
                logging.error(
                    f"Connection to Minecraft Server lost, "
                    f"polling terminated in {get_thread_name()}"
                )
                self._polling = False
            except SessionTimeout:
                logging.warning(f"Connection timeout in {get_thread_name()}")
            except PlayerNotInWorld as e:
                logging.warning(e)
                self._polling = False
            except BaseException:
                logging.error(f"Error in {get_thread_name()}", exc_info=True)

            if self.once:
                self._polling = False

        if self in self.monitors:
            self.monitors.remove(self)
        self.poll_thread = None
        self.pollers = []
        logging.info(f"Monitor {self.name} stopped")

    def add_poller_func(self, func: CallbackFunction):
        self.pollers.append(func)
        self._start_poller()

    def remove_poller_func(self, func: CallbackFunction):
        idx = self.pollers.index(func)
        if idx < 0:
            logging.error("removing unknown poller function")
        else:
            self.pollers.remove(func)

    @classmethod
    def stop_all(cls):
        for monitor in cls.monitors:
            monitor.stop()
        cls.monitors.clear()
        logging.info("Stopped all monitoring threads")

    def stop(self):
        self._polling = False

    def __del__(self):
        self.stop()

    def __repr__(self):
        # TODO work out how to shoe the class of the bound method
        func_list = [f.__name__ for f in self.pollers]
        return f"{self.name} polling {func_list} at {self.poll_rate})"



"""
Functions to manage threads.

Each thread maintains its own Client object for parallel execution of Minecraft
Server functions. The 1st Client object is created on the main thread and then
passed to new_thread.
"""

import threading

from mciwb import Client

thread_local = threading.local()


def new_thread(client: Client, target, name: str) -> threading.Thread:
    """
    Create a thread with its own RCON Client connection stored in thread local storage
    """

    def _enter_thread(client, target, name):
        # make new client connection for this thread
        new_client = type(client)(client.host, client.port, passwd=client.passwd)
        new_client.connect(True)

        # save our new client in the thread local storage
        set_client(new_client)
        thread_local.name = name

        target()

    # Start a new thread and use _enter_thread to set it up with a RCON client
    new_thread = threading.Thread(target=_enter_thread, args=(client, target, name))
    new_thread.start()

    return new_thread


def set_client(client: Client) -> None:
    # save our new client in the thread local storage
    thread_local.client = client


def get_client() -> Client:
    # retrieve the client for this thread
    return thread_local.client


def get_thread_name() -> str:
    return thread_local.name



from time import sleep

from mciwb import Monitor


def test_single_use(mock_client):
    def test_func():
        nonlocal count
        count += 1

    count = 0

    mon = Monitor(test_func, once=True)
    while mon._polling:
        sleep(0.1)

    assert count == 1


def test_multiple_use(mock_client):
    def test_func():
        nonlocal count
        count += 1

    count = 0

    mon = Monitor(test_func, poll_rate=0.0001)
    while count < 100:
        sleep(0.1)
    mon.remove_poller_func(test_func)

    Monitor.stop_all()

    assert count >= 100



import logging

from mcipc.rcon.enumerations import CloneMode, Item, MaskMode
from mcwb.types import Vec3

from mciwb.threads import get_client

zero = Vec3(0, 0, 0)


class CopyPaste:
    """
    Provides an interactive way to use copy and paste commands within a
    Minecraft world.
    """

    def __init__(self):
        self.start_pos: Vec3 = zero
        self.stop_pos: Vec3 = self.start_pos
        self.paste_pos: Vec3 = self.start_pos
        self._clone_dest = zero
        self.size = zero

    def get_commands(self):
        return {
            "select": self.select,
            "paste": self.paste,
            "expand": self.expand_to,
            "clear": self.clear,
        }

    def select(self, pos: Vec3):
        """
        Select a new copy buffer start point in the world.
        The previous start point becomes the stop point
        (i.e. opposite corner of the paste buffer)
        """
        self.stop_pos = self.start_pos
        self.start_pos = pos.with_ints()
        self.size = self.stop_pos - self.start_pos
        self._set_paste(Vec3(self.start_pos.x, self.start_pos.y, self.start_pos.z))

    def _set_paste(self, pos: Vec3):
        """
        Set the paste point relative to the current player position
        """
        self.paste_pos = pos
        # adjust clone dest so the paste corner matches the start paste buffer
        self._clone_dest = self.paste_pos
        x_off = self.size.x if self.size.x < 0 else 0
        y_off = self.size.y if self.size.y < 0 else 0
        z_off = self.size.z if self.size.z < 0 else 0
        self._clone_dest += Vec3(x_off, y_off, z_off)

    def paste(self, pos: Vec3, force=True):
        """
        Copy the contents of past buffer to position x y z
        """
        client = get_client()
        self._set_paste(pos)
        mode = CloneMode.FORCE if force else CloneMode.NORMAL
        result = client.clone(
            self.start_pos,
            self.stop_pos,
            self._clone_dest,
            mask_mode=MaskMode.REPLACE,
            clone_mode=mode,
        )
        logging.info(result)

    def paste_safe(self, pos: Vec3):
        self.paste(pos, force=False)

    def fill(self, pos: Vec3 = zero, item: Item = Item.AIR):
        """
        fill the paste buffer offset by x y z with Air or a specified block
        """
        client = get_client()

        offset = pos
        end = self.paste_pos + self.size + offset
        result = client.fill(self.paste_pos + offset, end, item.value)
        logging.info(result)

    def clear(self, _: Vec3 = zero):
        """
        Clear the current paste buffer
        """
        self.fill()

    def expand_to(self, pos: Vec3):
        """
        expand one or more of the dimensions of the copy buffer by moving
        the faces outwards to the specified point
        """
        # use mutable start and stop here to make the code more readable
        start = self.start_pos._asdict()
        stop = self.stop_pos._asdict()

        for dim in ["x", "y", "z"]:
            if start[dim] <= stop[dim]:
                if pos[dim] > stop[dim]:
                    stop[dim] = pos[dim]
                elif pos[dim] < start[dim]:
                    start[dim] = pos[dim]
            elif start[dim] >= stop[dim]:
                if pos[dim] < stop[dim]:
                    stop[dim] = pos[dim]
                elif pos[dim] > start[dim]:
                    start[dim] = pos[dim]

        self.select(Vec3(**stop))
        self.select(Vec3(**start))

    def expand(self, x=0, y=0, z=0):
        """
        expand one or more of the dimensions of the copy buffer by relative
        amounts
        """
        expander = Vec3(x, y, z)
        # use mutable start and stop here to make the code more readable
        start = self.start_pos._asdict()
        stop = self.stop_pos._asdict()

        for dim in ["x", "y", "z"]:
            if expander[dim] > 0:
                if start[dim] > stop[dim]:
                    start[dim] += expander[dim]
                else:
                    stop[dim] += expander[dim]
            elif expander[dim] < 0:
                if start[dim] < stop[dim]:
                    start[dim] += expander[dim]
                else:
                    stop[dim] += expander[dim]

        self.select(Vec3(**stop))
        self.select(Vec3(**start))



from mcipc.rcon.item import Item
from mcwb import Blocks, Volume
from mcwb.itemlists import grab
from mcwb.types import Anchor3, Cuboid, Vec3

from mciwb.threads import get_client


class SampleCube:
    def __init__(self) -> None:
        blue_row = [Item.BLUE_CONCRETE, Item.BLUE_CONCRETE, Item.BLUE_CONCRETE]
        red_row = [Item.RED_CONCRETE, Item.RED_CONCRETE, Item.RED_CONCRETE]
        hollow_row = [Item.RED_CONCRETE, Item.AIR, Item.RED_CONCRETE]
        top_profile = [blue_row, red_row, red_row]
        middle_profile = [blue_row, hollow_row, red_row]
        bottom_profile = [blue_row, blue_row, red_row]
        self.cube: Cuboid = [top_profile, middle_profile, bottom_profile]
        self.size = Vec3(2, 2, 2)
        # TODO does Volume's use of size need review?
        self.volume_size = Vec3(3, 3, 3)

        self.air = (((Item.AIR,) * 3) * 3) * 3

    def create(self, pos: Vec3, anchor: Anchor3 = Anchor3.BOTTOM_NW):
        """
        create a set of blocks in the world at pos
        """
        Blocks(get_client(), pos, self.cube, anchor)

    def test(self, pos: Vec3, anchor: Anchor3 = Anchor3.BOTTOM_NW):
        """
        test that the test cube blocks exist at pos
        """
        dest_vol = Volume.from_anchor(pos, self.volume_size, anchor)
        dest_cuboid = grab(get_client(), dest_vol)

        assert dest_cuboid == self.cube

    def clear(self, pos: Vec3, anchor: Anchor3 = Anchor3.BOTTOM_NW):
        """
        clear blocks at location
        """
        volume = Volume.from_anchor(pos, self.volume_size, anchor)
        volume.fill(get_client(), Item.AIR)



from mcwb.types import Vec3
from mcwb.volume import Volume

from mciwb.player import Player
from tests.conftest import ENTITY_NAME


def test_players_in(mock_client):
    """
    verify that the players_in function returns a list of players
    """
    v = Volume.from_corners(Vec3(-10, -60, -10), Vec3(10, -60, 10))
    players = Player.players_in(v)
    assert len(players) == 1
    assert players[0] == ENTITY_NAME

    v = Volume.from_corners(Vec3(3, 3, 3), Vec3(10, 10, 10))
    players = Player.players_in(v)
    assert len(players) == 0


def test_player_in(mock_client, minecraft_player: Player):
    """
    verify that the player_in function returns True if the player is in the volume
    """
    v = Volume.from_corners(Vec3(-10, -1, -10), Vec3(10, -60, 10))
    assert minecraft_player.player_in(v)

    v = Volume.from_corners(Vec3(3, 3, 3), Vec3(10, -60, 10))
    assert not minecraft_player.player_in(v)



import logging
from typing import Dict, Optional

from mcipc.rcon.exceptions import NoPlayerFound
from mcipc.rcon.item import Item
from mcipc.rcon.je import Client
from mcwb import Direction, Vec3, Volume
from mcwb.itemlists import grab
from rcon.exceptions import SessionTimeout

from mciwb.copier import CopyPaste
from mciwb.monitor import Monitor
from mciwb.player import Player, PlayerNotInWorld
from mciwb.signs import Signs
from mciwb.threads import get_client, set_client

world: "Iwb" = None  # type: ignore


class Iwb:
    """
    Interactive World Builder class. Provides a very simple interface for
    interactive functions for use in an IPython shell.
    """

    def __init__(self, server: str, port: int, passwd: str, client=None) -> None:
        self._server = server
        self._port = port
        self._passwd = passwd

        client = client or self.connect()

        self.player: Player = None  # type: ignore
        self.copier: CopyPaste = None  # type: ignore

        self._players: Dict[str, Player] = {}
        self._copiers: Dict[str, CopyPaste] = {}

    def connect(self) -> Client:
        """
        Makes a connection to the Minecraft Server. Can be called again
        if the connection is lost e.g. due to server reboot.
        """

        client = Client(self._server, int(self._port), passwd=self._passwd)
        client.connect(True)

        # store the client for the main thread
        set_client(client)

        logging.info(f"Connected to {self._server} on {self._port}")
        # don't announce every rcon command
        client.gamerule("sendCommandFeedback", False)

        return client

    def add_player(self, name: str, me=True):
        try:
            player = Player(name)
            self._players[name] = player

            sign = Signs(player)
            Monitor(sign.poll, name=name)
            self._copiers[name] = sign.copy

            if me:
                self.player = player
                self.copier = sign.copy

            sign.give_signs()
        except (PlayerNotInWorld, SessionTimeout, NoPlayerFound) as e:
            # during tests this will fail as there is no real player
            logging.warning("failed to give signs to player, %s", e)

        logging.info(f"Monitoring player {name} enabled for sign commands")

    def stop(self):
        Monitor.stop_all()

    def set_block(self, pos: Vec3, block: Item, facing: Optional[Vec3] = None):
        """
        Sets a block in the world
        """
        client = get_client()

        pos = Vec3(*pos)
        int_pos = pos.with_ints()
        nbt = []

        if facing:
            nbt.append(f"""facing={Direction.name(facing)}""")
        block_str = f"""{block}[{",".join(nbt)}]"""

        result = client.setblock(int_pos, block_str)
        logging.debug("setblock: " + result)

        # 'Could not set the block' is not an error - it means it was already set
        if not any(
            x in result for x in ["Changed the block", "Could not set the block", ""]
        ):
            logging.error(result)

    def get_block(self, pos: Vec3) -> Item:
        """
        Gets a block in the world
        """
        client = get_client()

        int_pos = Vec3(*pos).with_ints()

        grab_volume = Volume.from_corners(int_pos, int_pos)
        blocks = grab(client, grab_volume)

        return blocks[0][0][0]

    def __repr__(self) -> str:
        report = "Minecraft Interactive World Builder status:\n"
        if self.copier is not None:
            report += (
                "  copy buffer start: {o.copier.start_pos}\n"
                "  copy buffer stop: {o.copier.stop_pos}\n"
                "  copy buffer size: {o.copier.size}\n"
                "  paste point: {o.copier.paste_pos}\n"
            )
        if self.player is not None:
            report += (
                "  player: {o.player.name}\n"
                "  player position: {o.player.pos}\n"
                "  player facing: {o.player.facing}\n"
            )
        else:
            report += "  no player selected\n"

        return report.format(o=self)



import sys
from subprocess import check_output

from typer.testing import CliRunner

from mciwb import __version__

runner = CliRunner()


def test_cli_version():
    cmd = [sys.executable, "-m", "mciwb", "--version"]
    assert check_output(cmd).decode().strip() == __version__



import time

from demo.shapes import pyramid
from mciwb import Direction, Item, Position, world


def follow(pyramid_centre):
    """
    Spooky Pyramid function: draws a pyramid centred at pyramid_centre, made
    of carved pumpkins that always turn to face the player
    """
    pyramid_centre = Position(*pyramid_centre)
    size = 11
    half = int(size / 2)
    corner = pyramid_centre + Direction.SOUTH * half + Direction.WEST * half

    old_direction = 0
    while True:
        time.sleep(0.5)
        direction = Direction.facing(Position(5, 0, -5), world.player.pos)
        if direction != old_direction:
            old_direction = direction
            pyramid(corner, size, Item.CARVED_PUMPKIN, facing=direction)



import logging
from typing import Optional

import typer
from IPython.terminal.embed import InteractiveShellEmbed

import mciwb
from mciwb import Iwb, __version__

cli = typer.Typer()


def version_callback(value: bool):
    if value:
        typer.echo(__version__)
        raise typer.Exit()


def exception_handler(exception_type, exception, traceback):
    if logging.root.level > logging.DEBUG:
        logging.error("%s: %s", exception_type.__name__, exception)
    logging.debug("", exc_info=True)


@cli.command()
def main(
    server: str = typer.Option(..., prompt=True),
    port: int = typer.Option(..., prompt=True),
    passwd: str = typer.Option(..., prompt=True, hide_input=True),
    player: str = "",
    debug: bool = False,
    test: bool = False,
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Print the version of ibek and exit",
    ),
):
    if debug:
        logging.basicConfig(
            format="%(levelname)s: %(pathname)s:%(lineno)d %(funcName)s "
            "\n\t%(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(levelname)s:\t%(message)s",
            level=logging.INFO,
        )

    try:
        mciwb.world = Iwb(server, port, passwd)

        if player:
            mciwb.world.add_player(player)
    except BaseException:
        mciwb.world.stop()
        raise

    # for quick access in the shell without qualifying the namespace
    world = mciwb.world

    logging.info("######### Starting Interactive Session ##########\n")

    # Prepare IPython shell with auto-reload so user code edits work immediately
    shell = InteractiveShellEmbed()
    shell.magic(r"%load_ext autoreload")
    shell.magic(r"%autoreload 2")
    # Also suppress traceback to avoid intimidating novice programmers
    ipython = shell.get_ipython()
    ipython._showtraceback = exception_handler  # type: ignore

    if test:
        # for testing just report the world object state
        print(mciwb.world)
    else:
        # enter iPython shell until users exits
        shell(colors="neutral")

    # Terminate all threads after interactive session exits
    world.stop()


if __name__ == "__main__":
    cli()



from pathlib import Path

from typer.testing import CliRunner

from mciwb.__main__ import cli
from tests.conftest import HOST, RCON_P, RCON_PORT

runner = CliRunner()


def run_cli(*args):
    result = runner.invoke(cli, [str(x) for x in args])
    if result.exception:
        raise result.exception
    assert result.exit_code == 0, result
    return result


def test_repr(tmp_path: Path, minecraft_container, minecraft_client, minecraft_player):
    """launch the cli in test mode and connect to the test server"""
    result = run_cli(
        "--test", "--server", HOST, "--port", RCON_PORT, "--passwd", RCON_P
    )

    assert "no player selected" in result.stdout

    result = run_cli(
        "--test",
        "--server",
        HOST,
        "--port",
        RCON_PORT,
        "--passwd",
        RCON_P,
        "--player",
        "george",
    )

    assert "player: george" in result.stdout



from tests.test_unit.test_copier import copy_anchors


def test_copy_anchors(minecraft_client):
    """
    Runs the unit test copy_anchors() but uses a real client so tests
    against a real minecraft server instead of a mock one.
    """
    copy_anchors()



from typing import List

from mcwb.types import Vec3
from mcwb.volume import Volume

from mciwb import Direction
from tests.conftest import ENTITY_NAME
from tests.mockclient import MockClient


class MockPlayer:
    def __init__(self, client: MockClient, name: str) -> None:
        self.name = ENTITY_NAME

    @property
    def pos(self) -> Vec3:
        return Vec3(0, 0, 0)

    @property
    def facing(self) -> Vec3:
        return Direction.SOUTH

    def _pos(self, client=None) -> Vec3:
        return self._pos()

    def _facing(self, client=None) -> Vec3:
        return self._facing()

    def player_in(self, volume: Volume) -> bool:
        """
        Check if the player is in the volume
        """
        return volume.inside(self.pos)

    @classmethod
    def players_in(cls, volume: Volume) -> List["MockPlayer"]:
        """
        return a list of player names whose position is inside the volume
        """
        return []
